import assert from 'assert';
import { Buffer } from 'buffer';
import { Readable } from 'stream';
import cacache from 'cacache';
function getBodyAndMetaKeys(key) {
    return [`${key}body`, `${key}meta`];
}
export class FileSystemCache {
    constructor(options = {}) {
        this.ttl = options.ttl;
        this.cacheDirectory = options.cacheDirectory ?? '.cache';
    }
    async get(key, options) {
        const [, metaKey] = getBodyAndMetaKeys(key);
        const metaInfo = await cacache.get.info(this.cacheDirectory, metaKey);
        if (!metaInfo) {
            return undefined;
        }
        const metaBuffer = await cacache.get.byDigest(this.cacheDirectory, metaInfo.integrity);
        const metaData = JSON.parse(metaBuffer);
        const { bodyStreamIntegrity, empty, expiration } = metaData;
        delete metaData.bodyStreamIntegrity;
        delete metaData.empty;
        delete metaData.expiration;
        if (!options?.ignoreExpiration && expiration && expiration < Date.now()) {
            return undefined;
        }
        const bodyStream = Boolean(empty) || !bodyStreamIntegrity
            ? Readable.from(Buffer.alloc(0))
            : cacache.get.stream.byDigest(this.cacheDirectory, bodyStreamIntegrity);
        return {
            bodyStream,
            metaData,
        };
    }
    async remove(key) {
        const [bodyKey, metaKey] = getBodyAndMetaKeys(key);
        return Promise.all([
            cacache.rm.entry(this.cacheDirectory, bodyKey),
            cacache.rm.entry(this.cacheDirectory, metaKey),
        ]);
    }
    async set(key, bodyStream, metaData) {
        const [bodyKey, metaKey] = getBodyAndMetaKeys(key);
        const metaToStore = {
            ...metaData,
            expiration: undefined,
            bodyStreamIntegrity: undefined,
            empty: false,
        };
        if (typeof this.ttl === 'number') {
            metaToStore.expiration = Date.now() + this.ttl;
        }
        try {
            metaToStore.bodyStreamIntegrity = await new Promise((fulfill, reject) => {
                bodyStream.pipe(cacache.put.stream(this.cacheDirectory, bodyKey))
                    .on('integrity', (i) => {
                    fulfill(i);
                })
                    .on('error', (error) => {
                    reject(error);
                });
            });
        }
        catch (error) {
            if (error.code !== 'ENODATA') {
                throw error;
            }
            metaToStore.empty = true;
        }
        const metaBuffer = Buffer.from(JSON.stringify(metaToStore));
        await cacache.put(this.cacheDirectory, metaKey, metaBuffer);
        const cachedData = await this.get(key, { ignoreExpiration: true });
        assert(cachedData, 'Failed to cache response');
        return cachedData;
    }
}
//# sourceMappingURL=file_system_cache.js.map