import fs from 'fs';
import crypto from 'crypto';
import { Buffer } from 'buffer';
import { Request as NodeFetchRequest } from 'node-fetch';
import { FormData } from '../types.js';
export const CACHE_VERSION = 5;
function md5(string_) {
    return crypto.createHash('md5').update(string_).digest('hex');
}
function getFormDataCacheKeyJson(formData) {
    // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
    const cacheKey = { ...formData };
    const boundary = formData.getBoundary();
    delete cacheKey._boundary;
    const boundaryReplaceRegex = new RegExp(boundary, 'g');
    cacheKey._streams = cacheKey._streams.map(s => {
        if (typeof s === 'string') {
            return s.replace(boundaryReplaceRegex, '');
        }
        return s;
    });
    return cacheKey;
}
function getHeadersCacheKeyJson(headers) {
    return headers
        .map(([key, value]) => [key.toLowerCase(), value])
        .filter(([key, value]) => key !== 'cache-control' || value !== 'only-if-cached');
}
function getBodyCacheKeyJson(body) {
    if (!body) {
        return undefined;
    }
    if (typeof body === 'string') {
        return body;
    }
    if (body instanceof URLSearchParams) {
        return body.toString();
    }
    if (body instanceof fs.ReadStream) {
        return body.path.toString();
    }
    if (body instanceof FormData) {
        return getFormDataCacheKeyJson(body);
    }
    if (body instanceof Buffer) {
        return body.toString();
    }
    throw new Error('Unsupported body type. Supported body types are: string, number, undefined, null, url.URLSearchParams, fs.ReadStream, FormData');
}
function getRequestCacheKeyJson(request) {
    return {
        headers: getHeadersCacheKeyJson([...request.headers.entries()]),
        method: request.method,
        redirect: request.redirect,
        referrer: request.referrer,
        url: request.url,
        body: getBodyCacheKeyJson(request.body),
        follow: request.follow,
        compress: request.compress,
        size: request.size,
    };
}
export function calculateCacheKey(resource, init) {
    const resourceCacheKeyJson = resource instanceof NodeFetchRequest
        ? getRequestCacheKeyJson(resource)
        : { url: resource, body: undefined };
    const initCacheKeyJson = {
        body: undefined,
        ...init,
        headers: getHeadersCacheKeyJson(Object.entries(init?.headers ?? {})),
    };
    resourceCacheKeyJson.body = getBodyCacheKeyJson(resourceCacheKeyJson.body);
    initCacheKeyJson.body = getBodyCacheKeyJson(initCacheKeyJson.body);
    delete initCacheKeyJson.agent;
    return md5(JSON.stringify([resourceCacheKeyJson, initCacheKeyJson, CACHE_VERSION]));
}
//# sourceMappingURL=cache_keys.js.map