'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var fetch = require('node-fetch');
var FormData = require('form-data');
var locko = require('locko');
var assert = require('assert');
var buffer = require('buffer');
var stream = require('stream');
var fs = require('fs');
var crypto = require('crypto');
var cacache = require('cacache');

const responseInternalSymbol = Object.getOwnPropertySymbols(new fetch.Response())[1];
assert(responseInternalSymbol, 'Failed to get node-fetch responseInternalSymbol');
class NFCResponse extends fetch.Response {
    static serializeMetaFromNodeFetchResponse(response) {
        const metaData = {
            url: response.url,
            status: response.status,
            statusText: response.statusText,
            headers: response.headers.raw(),
            size: response.size,
            timeout: response.timeout,
            counter: response[responseInternalSymbol].counter,
        };
        return metaData;
    }
    static cacheMissResponse(url) {
        return new NFCResponse(stream.Readable.from(buffer.Buffer.alloc(0)), {
            url,
            status: 504,
            statusText: 'Gateway Timeout',
            headers: {},
            size: 0,
            timeout: 0,
            counter: 0,
        }, async () => undefined, false, true);
    }
    constructor(bodyStream, metaData, ejectFromCache, returnedFromCache, isCacheMiss = false) {
        // The node-fetch types seem to have a bug. response.headers.raw() returns a Record<string, string[]>
        // but the types claim that the constructor does not accept headers in that format (even though it
        // empirically does). So we cast it to unknown and then to the type that is accepted (even though it's
        // not really that type). I might go send the types a PR.
        super(bodyStream, metaData);
        this.ejectFromCache = ejectFromCache;
        this.returnedFromCache = returnedFromCache;
        this.isCacheMiss = isCacheMiss;
    }
}

class KeyTimeout {
    constructor() {
        this.timeoutHandleForKey = {};
    }
    clearTimeout(key) {
        clearTimeout(this.timeoutHandleForKey[key]);
    }
    updateTimeout(key, durationMs, callback) {
        this.clearTimeout(key);
        this.timeoutHandleForKey[key] = setTimeout(() => {
            callback();
        }, durationMs);
    }
}

async function streamToBuffer(stream) {
    const chunks = [];
    return new Promise((resolve, reject) => {
        stream.on('data', chunk => chunks.push(chunk)).on('error', error => {
            reject(error);
        }).on('end', () => {
            resolve(buffer.Buffer.concat(chunks));
        });
    });
}
class MemoryCache {
    constructor(options) {
        this.keyTimeout = new KeyTimeout();
        this.cache = new Map();
        this.ttl = options?.ttl;
    }
    async get(key) {
        const cachedValue = this.cache.get(key);
        if (cachedValue) {
            return {
                bodyStream: stream.Readable.from(cachedValue.bodyBuffer),
                metaData: cachedValue.metaData,
            };
        }
        return undefined;
    }
    async remove(key) {
        this.keyTimeout.clearTimeout(key);
        this.cache.delete(key);
    }
    async set(key, bodyStream, metaData) {
        const bodyBuffer = await streamToBuffer(bodyStream);
        this.cache.set(key, { bodyBuffer, metaData });
        if (typeof this.ttl === 'number') {
            this.keyTimeout.updateTimeout(key, this.ttl, async () => this.remove(key));
        }
        const cachedResult = await this.get(key);
        assert(cachedResult, 'Failed to cache response');
        return cachedResult;
    }
}

const CACHE_VERSION = 5;
function md5(string_) {
    return crypto.createHash('md5').update(string_).digest('hex');
}
function getFormDataCacheKeyJson(formData) {
    // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
    const cacheKey = { ...formData };
    const boundary = formData.getBoundary();
    delete cacheKey._boundary;
    const boundaryReplaceRegex = new RegExp(boundary, 'g');
    cacheKey._streams = cacheKey._streams.map(s => {
        if (typeof s === 'string') {
            return s.replace(boundaryReplaceRegex, '');
        }
        return s;
    });
    return cacheKey;
}
function getHeadersCacheKeyJson(headers) {
    return headers
        .map(([key, value]) => [key.toLowerCase(), value])
        .filter(([key, value]) => key !== 'cache-control' || value !== 'only-if-cached');
}
function getBodyCacheKeyJson(body) {
    if (!body) {
        return undefined;
    }
    if (typeof body === 'string') {
        return body;
    }
    if (body instanceof URLSearchParams) {
        return body.toString();
    }
    if (body instanceof fs.ReadStream) {
        return body.path.toString();
    }
    if (body instanceof FormData) {
        return getFormDataCacheKeyJson(body);
    }
    if (body instanceof buffer.Buffer) {
        return body.toString();
    }
    throw new Error('Unsupported body type. Supported body types are: string, number, undefined, null, url.URLSearchParams, fs.ReadStream, FormData');
}
function getRequestCacheKeyJson(request) {
    return {
        headers: getHeadersCacheKeyJson([...request.headers.entries()]),
        method: request.method,
        redirect: request.redirect,
        referrer: request.referrer,
        url: request.url,
        body: getBodyCacheKeyJson(request.body),
        follow: request.follow,
        compress: request.compress,
        size: request.size,
    };
}
function calculateCacheKey(resource, init) {
    const resourceCacheKeyJson = resource instanceof fetch.Request
        ? getRequestCacheKeyJson(resource)
        : { url: resource, body: undefined };
    const initCacheKeyJson = {
        body: undefined,
        ...init,
        headers: getHeadersCacheKeyJson(Object.entries(init?.headers ?? {})),
    };
    resourceCacheKeyJson.body = getBodyCacheKeyJson(resourceCacheKeyJson.body);
    initCacheKeyJson.body = getBodyCacheKeyJson(initCacheKeyJson.body);
    delete initCacheKeyJson.agent;
    return md5(JSON.stringify([resourceCacheKeyJson, initCacheKeyJson, CACHE_VERSION]));
}

const cacheOkayOnly = (response) => response.ok;
const cacheNon5xxOnly = (response) => response.status < 500;

function headerKeyIsCacheControl(key) {
    return key.trim().toLowerCase() === 'cache-control';
}
function headerValueContainsOnlyIfCached(cacheControlValue) {
    return cacheControlValue
        ?.split?.(',')
        .map(d => d.trim().toLowerCase())
        .includes('only-if-cached');
}
function headerEntryIsCacheControlOnlyIfCached(pair) {
    return headerKeyIsCacheControl(pair[0]) && headerValueContainsOnlyIfCached(pair[1]);
}
function hasOnlyIfCachedOption(resource, init) {
    const initHeaderEntries = Object.entries(init?.headers ?? {});
    const initHeaderEntriesContainsCacheControlOnlyIfCached = initHeaderEntries.some(pair => headerEntryIsCacheControlOnlyIfCached(pair));
    if (initHeaderEntriesContainsCacheControlOnlyIfCached) {
        return true;
    }
    if (resource instanceof fetch.Request
        && headerValueContainsOnlyIfCached(resource.headers.get('Cache-Control') ?? undefined)) {
        return true;
    }
    return false;
}

/* This is a bit of a hack to deal with the case when the user
 * consumes the response body in their `shouldCacheResponse` delegate.
 * The response body can only be consumed once, so if the user consumes
 * it then we wouldn't be able to read it again to write it to the cache.
 * This shim allows us to intercept the parsed body in that case and repackage
 * it into a fresh stream to cache. This of course doesn't work if the user
 * reads response.body directly, but that's not going to be likely.
 * My initial inclination was to use Response.prototype.clone() for this,
 * but the problems with backpressure seem significant. */
function shimResponseToSnipeBody(response, replaceBodyStream) {
    const origArrayBuffer = response.arrayBuffer;
    response.arrayBuffer = async function () {
        const arrayBuffer = await origArrayBuffer.call(this);
        replaceBodyStream(stream.Readable.from(buffer.Buffer.from(arrayBuffer)));
        return arrayBuffer;
    };
    const origBuffer = response.buffer;
    response.buffer = async function () {
        const buffer = await origBuffer.call(this);
        replaceBodyStream(stream.Readable.from(buffer));
        return buffer;
    };
    const origJson = response.json;
    response.json = async function () {
        const json = await origJson.call(this);
        replaceBodyStream(stream.Readable.from(buffer.Buffer.from(JSON.stringify(json))));
        return json;
    };
    const origText = response.text;
    response.text = async function () {
        const text = await origText.call(this);
        replaceBodyStream(stream.Readable.from(buffer.Buffer.from(text)));
        return text;
    };
    const origBlob = response.blob;
    response.blob = async function () {
        const blob = await origBlob.call(this);
        replaceBodyStream(stream.Readable.from(buffer.Buffer.from(await blob.text())));
        return blob;
    };
    const origTextConverted = response.textConverted;
    response.textConverted = async function () {
        const textConverted = await origTextConverted.call(this);
        replaceBodyStream(stream.Readable.from(buffer.Buffer.from(textConverted)));
        return textConverted;
    };
}

function getBodyAndMetaKeys(key) {
    return [`${key}body`, `${key}meta`];
}
class FileSystemCache {
    constructor(options = {}) {
        this.ttl = options.ttl;
        this.cacheDirectory = options.cacheDirectory ?? '.cache';
    }
    async get(key, options) {
        const [, metaKey] = getBodyAndMetaKeys(key);
        const metaInfo = await cacache.get.info(this.cacheDirectory, metaKey);
        if (!metaInfo) {
            return undefined;
        }
        const metaBuffer = await cacache.get.byDigest(this.cacheDirectory, metaInfo.integrity);
        const metaData = JSON.parse(metaBuffer);
        const { bodyStreamIntegrity, empty, expiration } = metaData;
        delete metaData.bodyStreamIntegrity;
        delete metaData.empty;
        delete metaData.expiration;
        if (!options?.ignoreExpiration && expiration && expiration < Date.now()) {
            return undefined;
        }
        const bodyStream = Boolean(empty) || !bodyStreamIntegrity
            ? stream.Readable.from(buffer.Buffer.alloc(0))
            : cacache.get.stream.byDigest(this.cacheDirectory, bodyStreamIntegrity);
        return {
            bodyStream,
            metaData,
        };
    }
    async remove(key) {
        const [bodyKey, metaKey] = getBodyAndMetaKeys(key);
        return Promise.all([
            cacache.rm.entry(this.cacheDirectory, bodyKey),
            cacache.rm.entry(this.cacheDirectory, metaKey),
        ]);
    }
    async set(key, bodyStream, metaData) {
        const [bodyKey, metaKey] = getBodyAndMetaKeys(key);
        const metaToStore = {
            ...metaData,
            expiration: undefined,
            bodyStreamIntegrity: undefined,
            empty: false,
        };
        if (typeof this.ttl === 'number') {
            metaToStore.expiration = Date.now() + this.ttl;
        }
        try {
            metaToStore.bodyStreamIntegrity = await new Promise((fulfill, reject) => {
                bodyStream.pipe(cacache.put.stream(this.cacheDirectory, bodyKey))
                    .on('integrity', (i) => {
                    fulfill(i);
                })
                    .on('error', (error) => {
                    reject(error);
                });
            });
        }
        catch (error) {
            if (error.code !== 'ENODATA') {
                throw error;
            }
            metaToStore.empty = true;
        }
        const metaBuffer = buffer.Buffer.from(JSON.stringify(metaToStore));
        await cacache.put(this.cacheDirectory, metaKey, metaBuffer);
        const cachedData = await this.get(key, { ignoreExpiration: true });
        assert(cachedData, 'Failed to cache response');
        return cachedData;
    }
}

function getUrlFromRequestArguments(resource) {
    if (resource instanceof fetch.Request) {
        return resource.url;
    }
    return resource;
}
async function getResponse(fetchCustomization, resource, init) {
    if (typeof resource !== 'string' && !(resource instanceof fetch.Request)) {
        throw new TypeError('The first argument to fetch must be either a string or a node-fetch Request instance');
    }
    const cacheKey = fetchCustomization.calculateCacheKey(resource, init);
    const ejectSelfFromCache = async () => fetchCustomization.cache.remove(cacheKey);
    return locko.doWithLock(cacheKey, async () => {
        const cachedValue = await fetchCustomization.cache.get(cacheKey);
        if (cachedValue) {
            return new NFCResponse(cachedValue.bodyStream, cachedValue.metaData, ejectSelfFromCache, true);
        }
        if (hasOnlyIfCachedOption(resource, init)) {
            return NFCResponse.cacheMissResponse(getUrlFromRequestArguments(resource));
        }
        const fetchResponse = await fetch(resource, init);
        const serializedMeta = NFCResponse.serializeMetaFromNodeFetchResponse(fetchResponse);
        let bodyStream = fetchResponse.body;
        shimResponseToSnipeBody(fetchResponse, stream => {
            bodyStream = stream;
        });
        const shouldCache = await fetchCustomization.shouldCacheResponse(fetchResponse);
        if (shouldCache) {
            const cacheSetResult = await fetchCustomization.cache.set(cacheKey, bodyStream, serializedMeta);
            bodyStream = cacheSetResult.bodyStream;
        }
        return new NFCResponse(bodyStream, serializedMeta, ejectSelfFromCache, false);
    });
}
const globalMemoryCache = new MemoryCache();
function create(creationOptions) {
    const fetchOptions = {
        cache: creationOptions.cache ?? globalMemoryCache,
        shouldCacheResponse: creationOptions.shouldCacheResponse ?? (() => true),
        calculateCacheKey: creationOptions.calculateCacheKey ?? calculateCacheKey,
    };
    const fetchCache = async (resource, init, perRequestOptions) => getResponse({ ...fetchOptions, ...perRequestOptions }, resource, init);
    fetchCache.create = create;
    fetchCache.options = fetchOptions;
    return fetchCache;
}
const defaultFetch = create({});
const cacheStrategies = {
    cacheOkayOnly,
    cacheNon5xxOnly,
};

Object.defineProperty(exports, "NodeFetchRequest", {
    enumerable: true,
    get: function () { return fetch.Request; }
});
exports.FormData = FormData;
exports.CACHE_VERSION = CACHE_VERSION;
exports.FileSystemCache = FileSystemCache;
exports.MemoryCache = MemoryCache;
exports.NodeFetchCache = defaultFetch;
exports.cacheStrategies = cacheStrategies;
exports.calculateCacheKey = calculateCacheKey;
exports.default = defaultFetch;
exports.getCacheKey = calculateCacheKey;
